"""
"""


#[
from __future__ import annotations

import json as _js
import copy as _co
import types as _ty
import numpy as _np
import re as _re
import operator as _op
import functools as _ft
import itertools as _it
import os as _os
from typing import (Self, TypeAlias, Literal, Sequence, Protocol, Any, )
from collections.abc import (Iterable, Iterator, Callable, )
from numbers import (Number, )

from ..conveniences import views as _views
from ..conveniences import descriptions as _descriptions
from ..conveniences import iterators as _iterators
from ..series import main as _series
from ..dataslates import main as _dataslates
from .. import quantities as _quantities
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import pages as _pages

from . import _imports as _imports
from . import _exports as _exports
from . import _merge as _merge
from . import _dotters as _dotters
from . import _fred as _fred
from . import _views as _views
#]


__all__ = (
    "Databox",
    "Databank",
)


SourceNames: TypeAlias = Iterable[str] | str | Callable[[str], bool] | None
TargetNames: TypeAlias = Iterable[str] | str | Callable[[str], str] | None
InterpretRange: TypeAlias = Literal["base", "extended", ]


class SteadyDataboxableProtocol(Protocol):
    """
    """
    #[

    def get_min_max_shifts(self, *args) -> tuple(int, int): ...
    def generate_steady_items(self, *args) -> Any: ...

    #]


def _extended_range_tuple_from_base_span(
    input_span: Iterable[_dates.Dater],
    min_shift: int,
    max_shift: int,
    prepend_initial: bool,
    append_terminal: bool,
    /,
) -> tuple[_dates.Date, _dates.Date]:
    """
    """
    range_list = tuple(t for t in input_span)
    start_date, end_date = range_list[0], range_list[-1]
    start_date += min_shift if prepend_initial else 0
    end_date += max_shift if append_terminal else 0
    return start_date, end_date


@_pages.reference(
    path=("data_management", "databoxes.md", ),
    categories={
        "constructor": "Creating new databoxes",
        "import_export": "Importing and exporting databoxes",
    },
)
class Databox(
    _imports.Inlay,
    _exports.Inlay,
    _merge.Inlay,
    _dotters.DotterMixin,
    _fred.FredMixin,
    _descriptions.DescriptionMixin,
    _views.ViewMixin,
    dict,
):
    """
......................................................................

Databoxes
==========

The `Databox` objects can be used to store and manipulate unstructured data
organized as key-value pairs, in a dictionary style; the `Databox` objects
are, indeed, subclassed from the standard `dict`. The values stored within
a databox can be of any type.

......................................................................
    """
    #[

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        """
        super().__init__(*args, **kwargs, )
        self._description = ""
        self._dotters = []

    @classmethod
    @_pages.reference(
        category="constructor",
        call_name="Databox.empty",
    )
    def empty(
        klass,
        /,
    ) -> Self:
        """
......................................................................

==Create a new empty `Databox`==

......................................................................
        """
        return klass()

    @classmethod
    @_pages.reference(category="constructor", call_name="Databox.from_dict", )
    def from_dict(
        klass,
        _dict: dict,
        /,
    ) -> Self:
        """
......................................................................

==Create a `Databox` from a `dict`==

......................................................................
        """
        self = klass()
        for k, v in _dict.items():
            self[k] = v
        return self

    @classmethod
    @_pages.reference(category="constructor", call_name="Databox.from_dict", )
    def from_array(
        klass,
        array: _np.ndarray,
        names: Sequence[str],
        *,
        descriptions: Iterable[str] | None = None,
        dates: Iterable[_dates.Date] | None = None,
        start_date: _dates.Date | None = None,
        target_databox: Self | None = None,
        orientation: Literal["vertical", "horizontal", ] = "vertical",
    ) -> Self:
        """
......................................................................

==Create a `Databox` from a NumPy array==

......................................................................
        """
        array = array if orientation == "horizontal" else array.T
        series_constructor = _get_series_constructor(start_date, dates, )
        return klass._from_horizontal_array_and_constructor(
            array,
            names,
            series_constructor,
            descriptions=descriptions,
            target_databox=target_databox,
        )

    @classmethod
    def _from_horizontal_array_and_constructor(
        klass,
        array: _np.ndarray,
        names: Iterable[str],
        series_constructor: Callable,
        *,
        descriptions: Sequence[str] | None = None,
        target_databox: Self | None = None,
    ) -> Self:
        """
        """
        self = target_databox or klass()
        descriptions = (
            descriptions if descriptions is not None
            else _it.repeat("", )
        )
        for name, values, description in zip(names, array, descriptions, ):
            print(name, type(values), description, )
            # self[name] = series_constructor(values=values, description=description, )
        return self

    def iter_variants(
        self,
        /,
        *,
        item_iterator: Iterator[Any] | None = None,
        names: Iterable[str] | None = None,
    ) -> Iterator[dict]:
        """
        """
        names = names or self.keys()
        item_iterator = item_iterator or _default_item_iterator
        dict_variant_iter = {
            k: item_iterator(self[k], )
            for k in names if k in self
        }
        while True:
            yield { k: next(v, ) for k, v in dict_variant_iter.items() }

    def get_names(self, /, ) -> list[str]:
        """
        Get all names stored in a databox save for private attributes
        """
        return tuple(self.keys())

    def get_missing_names(self, names: Iterable[str], /, ) -> tuple[str]:
        """
        Get names that are not in the databox
        """
        return tuple(name for name in names if name not in self)

    def get_num_items(self, /, ) -> int:
        """
        """
        return len(self.keys())

    @property
    @_pages.reference(category="property", )
    def num_items(self, /, ) -> int:
        """==Number of items in the databox=="""
        return len(self.keys())

    def to_dict(self: Self) -> dict:
        """
        Convert Databox to dict
        """
        return { k: v for k, v in self.items() }

    def copy(
        self: Self,
        /,
        source_names: SourceNames = None,
        target_names: TargetNames = None,
        strict_names: bool = False,
    ) -> Self:
        """
        """
        new_databox = _co.deepcopy(self)
        if source_names is None and target_names is None:
            return new_databox
        source_names, target_names, *_ = self._resolve_source_target_names(
            source_names, target_names, strict_names,
        )
        new_databox.rename(source_names, target_names, )
        new_databox.keep(target_names, )
        return new_databox

    def shallow(
        self: Self,
        /,
        source_names: SourceNames = None,
        target_names: TargetNames = None,
        strict_names: bool = False,
    ) -> Self:
        """
        """
        source_names, target_names, *_ = self._resolve_source_target_names(
            source_names, target_names, strict_names,
        )
        return type(self)(
            (t, self[s])
            for s, t in zip(source_names, target_names, )
        )

    def rename(
        self: Self,
        /,
        source_names: SourceNames = None,
        target_names: TargetNames = None,
        strict_names: bool = False,
    ) -> None:
        """
        """
        source_names, target_names, *_ = self._resolve_source_target_names(
            source_names, target_names, strict_names,
        )
        for s, t in zip(source_names, target_names, ):
            if s == t:
                continue
            self[t] = self.pop(s)

    def remove(
        self: Self,
        /,
        remove_names: SourceNames = None,
        strict_names: bool = False,
    ) -> None:
        """
        """
        if remove_names is None:
            return self
        context_names = self.get_names()
        remove_names, *_ = self._resolve_source_target_names(
            remove_names, None, strict_names,
        )
        for n in remove_names:
            del self[n]

    def keep(
        self: Self,
        /,
        keep_names: SourceNames = None,
        strict_names: bool = False,
    ) -> None:
        """
        """
        if keep_names is None:
            return self
        keep_names, _, context_names = self._resolve_source_target_names(
            keep_names, None, strict_names,
        )
        for n in context_names:
            if n in keep_names:
                continue
            del self[n]

    def apply(
        self,
        func: Callable,
        /,
        source_names: SourceNames = None,
        in_place: bool = True,
        when_fails: Literal["critical", "error", "warning", "silent", ] = "critical",
        strict_names: bool = False,
    ) -> None:
        """
        """
        source_names, *_ = self._resolve_source_target_names(
            source_names, None, strict_names,
        )
        when_fails_stream = \
            _wrongdoings.STREAM_FACTORY[when_fails] \
            (f"Error(s) when applying function to Databox items:")
        for s in source_names:
            try:
                output = func(self[s])
                if not in_place:
                    self[s] = output
            except Exception as e:
                when_fails_stream.add(f"{s}: {repr(e)}", )
        when_fails_stream._raise()

    def filter(
        self,
        /,
        name_test: Callable | None = None,
        value_test: Callable | None = None,
    ) -> Iterable[str]:
        """
        """
        names = tuple(self.get_names())
        if name_test is None and value_test is None:
            return names
        name_test = name_test if name_test else lambda x: True
        value_test = value_test if value_test else lambda x: True
        return tuple(
            name
            for name in names
            if name_test(name) and value_test(self[name], )
        )

    def get_series_names_by_frequency(
        self,
        frequency: _dates.Frequency,
    ) -> Iterable[str]:
        """
        Get all time series names with the given frequency
        """
        def _is_series_with_frequency(x):
            return isinstance(x, _series.Series) and x.frequency == frequency
        return self.filter(value_test=_is_series_with_frequency, )

    def get_span_by_frequency(
        self,
        frequency: _dates.Frequency,
    ) -> _dates.Ranger:
        """
        Get the encompassing date range for all time series with the given frequency
        """
        if frequency == _dates.Frequency.UNKNOWN:
            return _dates.EmptyRanger()
        names = self.get_series_names_by_frequency(frequency)
        if not names:
            return _dates.EmptyRanger()
        start_dates = (self[n].start_date for n in names)
        end_dates = (self[n].end_date for n in names)
        min_start_date = min(start_dates, key=_op.attrgetter("serial"), )
        max_end_date = max(end_dates, key=_op.attrgetter("serial"), )
        return _dates.Ranger(min_start_date, max_end_date, )

    def to_json(self, file_name, **kwargs):
        """
        """
        with open(file_name, "wt+") as f:
            return _js.dump(self, f, **kwargs)

    def overlay(
        self,
        other: Self,
        /,
        **kwargs,
    ) -> None:
        """
        """
        self._lay(other, _series.Series.overlay, **kwargs)

    def underlay(
        self,
        other: Self,
        /,
        **kwargs,
    ) -> None:
        """
        """
        self._lay(other, _series.Series.underlay, **kwargs)

    def _lay(
        self,
        other: Self,
        func: Callable,
        /,
        names: Iterable[str] | None = None,
        **kwargs,
    ) -> None:
        """"
        """
        if names is None:
            value_test = lambda x: isinstance(x, _series.Series)
            self_names = self.filter(value_test=value_test, )
            other_names = other.filter(value_test=value_test, )
            names = set(self_names).intersection(other_names)
        for n in names:
            if self[n].frequency == other[n].frequency:
                func(self[n], other[n], **kwargs, )

    def clip(
        self,
        /,
        new_start_date: _dates.Dater | None = None,
        new_end_date: _dates.Dater | None = None,
    ) -> None:
        """
        """
        if new_start_date is None and new_end_date is None:
            return
        frequency = (
            new_start_date.frequency
            if new_start_date is not None
            else new_end_date.frequency
        )
        value_test = lambda x: isinstance(x, _series.Series) and x.frequency == frequency
        names = self.filter(value_test=value_test, )
        for n in names:
            self[n].clip(new_start_date, new_end_date, )

    def prepend(
        self,
        prepending: Self,
        end_prepending: _dates.Dater,
        /,
    ) -> Self:
        """
        """
        prepending = prepending.copy()
        prepending.clip(None, end_prepending, )
        self.underlay(prepending, )

    def evaluate_expression(
        self,
        expression: str,
        /,
        context: dict[str, Any] | None = None,
    ) -> Any:
        """
        """
        expression = expression.strip()
        if expression in self:
            return self[expression]
        else:
            return self.eval(expression, context, )

    def eval(
        self,
        expression: str,
        /,
        context: dict[str, Any] | None = None,
    ) -> Any:
        """
        """
        expression = _reformat_eval_expression(expression, )
        globals = { k: v for k, v in self.items() }
        if context:
            globals |= context
        return eval(expression, globals, )

    def __call__(
        self,
        expression: str,
        /,
        context: dict[str, Any] | None = None,
    ) -> Any:
        """
        """
        return self.eval(expression, context, )

    @classmethod
    def steady(
        klass,
        steady_databoxable: SteadyDataboxableProtocol,
        input_span: Iterable[_dates.Dater],
        /,
        deviation: bool = False,
        prepend_initial: bool = True,
        append_terminal: bool = True,
    ) -> Self:
        """
        """
        self = klass()
        min_shift, max_shift = steady_databoxable.get_min_max_shifts()
        start_date, end_date = _extended_range_tuple_from_base_span(
            input_span,
            min_shift,
            max_shift,
            prepend_initial,
            append_terminal,
        )
        items = steady_databoxable.generate_steady_items(start_date, end_date, deviation=deviation, )
        for name, value in items:
            self[name] = value
        return self

    zero = _ft.partialmethod(steady, deviation=True, )

    def minus_control(
        self,
        model,
        control: Self,
        /,
    ) -> None:
        """
        """
        MINUS_FACTORY = {
            True: lambda x, y: x / y,
            False: lambda x, y: x - y,
            None: lambda x, y: x - y,
        }
        kind = _quantities.QuantityKind.ANY_VARIABLE | _quantities.QuantityKind.ANY_SHOCK
        quantities = model.get_quantities(kind=kind, )
        qid_to_name = _quantities.create_qid_to_name(quantities, )
        qid_to_logly = _quantities.create_qid_to_logly(quantities, )
        for qid, name in qid_to_name.items():
            minus_func = MINUS_FACTORY[qid_to_logly[qid]]
            self[name] = minus_func(self[name], control[name], )

    def __or__(self, other) -> Self:
        new = _co.deepcopy(self)
        new.update(other, )
        return new

    def _resolve_source_target_names(
        self,
        /,
        source_names: SourceNames,
        target_names: TargetNames,
        strict_names: bool = False,
    ) -> tuple[tuple[str, ...], tuple[str, ...], tuple[str, ...]]:
        """
        """
        context_names = self.get_names()
        if source_names is None:
            source_names = context_names
        if isinstance(source_names, str):
            source_names = tuple(source_names)
        if callable(source_names):
            func = source_names
            source_names = tuple(n for n in context_names if func(n))
        if target_names is None:
            target_names = source_names
        if isinstance(target_names, str):
            target_names = tuple(target_names)
        if callable(target_names):
            func = target_names
            target_names = tuple(func(n) for n in source_names)
        if not strict_names:
            source_target_pairs = tuple((s, t) for s, t in zip(source_names, target_names, ) if s in context_names)
            source_names, target_names = zip(*source_target_pairs, ) if source_target_pairs else ((), (), )
        return source_names, target_names, context_names

    #]


Databank = Databox


def _apply_to_item(
    func: Callable,
    source: Any,
    target: Any,
    /,
) -> None:
    """
    Apply function to source, capture result in target
    """
    target = source
    return func(target)


def _reformat_eval_expression(expression: str, ) -> str:
    """
    """
    #[
    if "=" in expression:
        lhs, *rhs = expression.split(expression, "=", )
        rhs = "=".join(rhs, )
        expression = "({lhs})-({rhs})"
    return expression
    #]


def _default_item_iterator(value: Any, /, ) -> Iterator[Any]:
    """
    """
    #[
    is_value_iterable = (
        isinstance(value, Iterable)
        and not isinstance(value, str)
        and not isinstance(value, bytes)
    )
    value = value if is_value_iterable else [value, ]
    yield from _iterators.exhaust_then_last(value, None, )
    #]


def _get_series_constructor(
    start_date: _dates.Dater | None = None,
    dates: Iterable[_dates.Dater] | None = None,
    /,
) -> Callable:
    """
    """
    #[
    if start_date is not None:
        return _ft.partial(
            _series.Series.from_start_date_and_values,
            start_date=start_date,
        )
    elif dates is not None:
        return _ft.partial(
            _series.Series.from_dates_and_values,
            dates=dates,
        )
    #]




"""
Assign custom values to quantities
"""


#[
from __future__ import annotations

from typing import (Any, )
from collections.abc import (Iterable, Iterator, )
from numbers import (Number, )

from ..conveniences import iterators as _iterators
from ..databoxes import main as _databoxes
#]


class AssignMixin:
    """
    """
    #[

    def assign(
        self,
        /,
        *args,
        **kwargs,
    ) -> None:
        """
        Assign parameters from dicts or from keyword arguments
        """
        for arg in args:
            self.assign(**arg, )
        if not kwargs:
            return
        name_to_qid = self.create_name_to_qid()
        qid_to_name = self.create_qid_to_name()
        qid_to_custom_values = _rekey_dict(kwargs, name_to_qid, )
        qid_to_custom_values_iter = _databoxes.Databox.iter_variants(
            qid_to_custom_values,
            item_iterator=_prepare_custom_value_iter,
        )
        for variant, values in zip(self._variants, qid_to_custom_values_iter, ):
            variant.update_values_from_dict(values, )
            self._enforce_assignment_rules(variant, )

    #]


def _rekey_dict(
    dict_to_rekey: dict,
    old_key_to_new_key: dict,
    /,
    garbage_key=None,
) -> dict[int, Any]:
    """
    """
    #[
    new_dict = {
        old_key_to_new_key.get(key, garbage_key): value
        for key, value in dict_to_rekey.items()
    }
    if garbage_key in new_dict:
        del new_dict[garbage_key]
    return new_dict
    #]


_UNCHANGED_VALUE = (..., ..., )


def _prepare_custom_value_iter(
    value: Any,
    /,
) -> Iterator[tuple[Number | EllipsisType, Number | EllipsisType]]:
    """
    Resolve custom values: value = [a, b, c] means variants, a = (1, 2) means level and change, ... means keep unchanged
    """
    #[
    is_iterable = (
        isinstance(value, Iterable, )
        and not isinstance(value, str, )
        and not isinstance(value, bytes, )
        and not isinstance(value, tuple, )
    )
    value = value if is_iterable else [value, ]
    return _iterators.exhaust_then_last(value, _UNCHANGED_VALUE, )
    #]




"""
"""


#[
from __future__ import annotations

import numpy as _np

from .. import quantities as _quantities
from .. import pages as _pages
from .. import wrongdoings as _wrongdoings
from ..databoxes import main as _databoxes
#]


class Inlay:
    """
    """
    #[

    @_pages.reference(category="parameters", )
    def assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
················································································

==Assign model parameters==

```
self.assign(
    name_one=value_one,
    name_two=value_two,
    # etc
)
```

```
self.assign(databox, )
```

...


### Input arguments ###


???+ input "self"

    `Sequential` model whose parameters will be assigned.


???+ input "name_one"

    Name of a parameter to assign.


???+ input "value_one"

    Value to assign to `name_one`.

etc...

???+ input "databox"

    `Databox` or `dict` from which the parameters will be extracted and
    assigned.

················································································
        """
        kwargs_to_assign = {}
        for a in args:
            kwargs_to_assign.update(a, )
        kwargs_to_assign.update(kwargs, )
        #
        kwargs_to_assign = {
            k: v
            for k, v in kwargs_to_assign.items()
            if k in self._invariant.parameter_names
        }
        #
        for v in self._variants:
            v.assign_from_dict_like(kwargs_to_assign, )

    def check_missing_parameters(self, /, ) -> _databoxes.Databox:
        """
        """
        parameters = self.get_parameters()
        missing = tuple(
            k for k, v in parameters.items()
            if _np.isnan(_np.array(v, dtype=float)).any()
        )
        if missing:
            raise _wrongdoings.IrisPieCritical(("Missing parameters", ) + missing)

    def get_parameters(self, /, ) -> _databoxes.Databox:
        """
        """
        parameter_names = self._invariant.parameter_names
        if self.is_singleton:
            return _databoxes.Databox(**{
                n: self._variants[0].parameters[n]
                for n in parameter_names
            })
        else:
            return _databoxes.Databox(**{
                n: [ v.parameters[n] for v in self._variants ]
                for n in parameter_names
            })

    #]




"""
"""


#[
from __future__ import annotations

from typing import (Any, Self, Iterable, Iterator, TypeVar, Protocol, )
from numbers import (Number, )
import numpy as _np

from .conveniences import iterators as _iterators
#]


_T = TypeVar("T")


class HasVariantsProtocol:
    """
    """
    #[

    _invariant: Any
    _variants: list

    #]


class HasVariantsMixin:
    """
    Mixin for handling multiple alternative variants of a single model object
    """
    #[

    @property
    def num_variants(self, /, ) -> int:
        """
        Number of alternative variants within this model
        """
        return len(self._variants)

    @property
    def is_singleton(self, /, ) -> bool:
        """
        True for models with only one variant
        """
        return is_singleton(self.num_variants, )

    def new_with_shallow_variants(self, /, ) -> Self:
        """
        Create a new model with pointers to invariant and variants of this model
        """
        new = type(self)()
        new._invariant = self._invariant
        new._variants = [ v for v in self._variants ]
        return new

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        Alter (expand, shrink) the number of alternative parameter variants in this model object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )

    def shrink_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        if new_num<1:
            raise Exception('Number of variants must be one or more')
        self._variants = (
            self._variants[0:new_num]
            if new_num < self.num_variants
            else self._variants
        )

    def expand_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        for i in range(self.num_variants, new_num):
            self._variants.append(self._variants[-1].copy(), )

    def broadcast_variants(self, other: Self, /, ) -> None:
        """
        """
        if self.num_variants < other.num_variants:
            self.expand_num_variants(other.num_variants, )

    def get_variant(
        self,
        vids: Iterable[int] | int | slice | EllipsisType,
        /,
    ) -> Self:
        new = type(self).skeleton(self, )
        new._invariant = self._invariant
        variant_iter = _resolve_vids(self, vids, )
        new._variants = [ self._variants[i] for i in variant_iter ]
        return new

    def iter_variants(
        self,
        /,
    ) -> Iterator[Self]:
        """
        """
        return _iterators.exhaust_then_last(self, )

    def iter_extract_variants(self, /, ) -> Iterator:
        """
        """
        return _iterators.exhaust_then_last(self._variants, )

    def _new_with_single_variant(self: _T, variant, /, ) -> _T:
        """
        """
        new = type(self).skeleton(self, )
        new._variants = [ variant ]
        return new

    def iter_own_variants(
        self,
        /,
    ) -> Iterator[Self]:
        """
        Iterate over alternative variants of this object
        """
        for v in self._variants:
            yield self._new_with_single_variant(v, )

    def __iter__(
        self,
        /,
    ) -> Iterator[Self]:
        """
        Iterate over alternative variants of this object
        """
        return self.iter_own_variants()

    def repack_singleton(
        self: HasVariantsProtocol,
        anything: list[_T] | _T,
        /,
    ) -> list[_T]:
        """
        """
        return repack_singleton(anything, self.is_singleton, )

    def unpack_singleton(
        self: HasVariantsProtocol,
        anything: list[_T] | _T,
        **kwargs,
    ) -> list[_T] | _T:
        """
        """
        return unpack_singleton(
            anything,
            self.is_singleton,
            **kwargs,
        )

    def unpack_singleton_in_dict(
        self: HasVariantsProtocol,
        anything_in_dict: dict[str, list[_T] | _T],
        **kwargs,
    ) -> dict[str, _T | list[_T]]:
        """
        """
        return unpack_singleton_in_dict(
            anything_in_dict,
            self.is_singleton,
            **kwargs,
        )
    #]


def _resolve_vids(
    self,
    vids: Iterable[int] | int | slice | EllipsisType,
    /,
) -> Iterable[int]:
    """
    """
    #[
    if isinstance(vids, Number):
        return (int(vids), )
    elif vids is ...:
        return range(self.num_variants, )
    elif isinstance(vids, slice):
        return range(*vids.indices(self.num_variants, ), )
    else:
        return tuple(int(i) for i in vids)
    #]


def iter_variants(anything: _T, ) -> Iterator[_T]:
    """
    """
    if hasattr(anything, "iter_variants", ):
        return anything.iter_variants()
    elif isinstance(anything, _np.ndarray, ) and anything.ndim:
        return _iterators.exhaust_then_last(anything.reshape(anything.shape[0], -1, ).T)
    elif isinstance(anything, list, ):
        return _iterators.exhaust_then_last(anything, )
    else:
        return _iterators.exhaust_then_last([anything, ], )


def iter_own_variants(anything: _T, ) -> Iterator[_T]:
    """
    """
    if hasattr(anything, "iter_own_variants", ):
        return anything.iter_own_variants()
    elif isinstance(anything, _np.ndarray, ) and anything.ndim:
        return anything.reshape(anything.shape[0], -1, ).T
    elif isinstance(anything, list, ):
        return anything
    else:
        return [anything, ]


def repack_singleton(anything: _T, is_singleton: bool, ) -> list[_T]:
    """
    """
    return [anything] if is_singleton else anything


def unpack_singleton(
    anything: list[_T],
    is_singleton: bool = True,
    /,
    unpack_singleton: bool = True,
) -> _T | list[_T]:
    """
    """
    return (
        anything[0]
        if is_singleton and unpack_singleton
        else anything
    )


shadow_unpack_singleton = unpack_singleton


def unpack_singleton_in_dict(
    anything_in_dict: dict[str, list[_T] | _T],
    is_singleton: bool,
    /,
    unpack_singleton: bool = True,
) -> dict[str, _T | list[_T]]:
    """
    """
    if is_singleton and unpack_singleton:
        return {
            k: shadow_unpack_singleton(v, )
            for k, v in anything_in_dict.items()
        }
    else:
        return anything_in_dict


def is_singleton(num_variants: int) -> bool:
    """
    """
    return num_variants == 1




"""
"""


#[
from __future__ import annotations

from typing import (Self, )

from . import main as _databoxes
#]


class Dotter:
    """
    """
    #[

    @classmethod
    def from_databox(
        klass,
        db: _databoxes.Databox,
    ) -> Self:
        """
        """
        self = klass()
        self.__dict__.update(db, )
        return self

    def to_databox(
        self,
        db: _databoxes.Databox,
    ) -> None:
        """
        """
        db.update(self.__dict__, )

    #]


class DotterMixin:
    """
    """
    #[

    def __enter__(
        self,
    ) -> Self:
        """
        """
        new_dotter = Dotter.from_databox(self, )
        self._dotters.append(new_dotter, )
        return new_dotter


    def __exit__(
        self,
        exc_type,
        exc_value,
        traceback,
    ) -> None:
        """
        """
        current_dotter = self._dotters.pop()
        current_dotter.to_databox(self, )

    #]




"""
"""


#[
from __future__ import annotations

from typing import (Self, Callable, NoReturn, )
from numbers import (Number, )
import copy as _cp

from ..conveniences import descriptions as _descriptions
from ..fords import descriptors as _descriptors
from ..equators import plain as _equators
from .. import equations as _equations
from .. import quantities as _quantities
from .. import wrongdoings as _wrongdoings
from .. import makers as _makers

from . import _flags
#]


_DEFAULT_STD_LINEAR = 1
_DEFAULT_STD_NONLINEAR = 0.01


_PLAIN_EQUATOR_EQUATION = (
    _equations.EquationKind.TRANSITION_EQUATION
    | _equations.EquationKind.MEASUREMENT_EQUATION
)


class Invariant(
    _descriptions.DescriptionMixin,
):
    """
    Invariant part of a Simultaenous object
    """
    #[

    _DEFAULT_STD_NAME_FORMAT = "std_{}"
    _DEFAULT_STD_DESCRIPTION_FORMAT = "(Std) {}"

    __slots__ = (
        "quantities",
        "dynamic_equations",
        "steady_equations",
        "dynamic_descriptor",
        "steady_descriptor",
        "shock_qid_to_std_qid",
        "_flags",
        "_context",
        "_plain_dynamic_equator",
        "_plain_steady_equator",
        "_min_shift",
        "_max_shift",
        "_default_std",
        "_description",
    )

    def __init__(
        self,
        source,
        /,
        check_syntax: bool = True,
        std_name_format: str | None = None,
        std_description_format: str | None = None,
        autodeclare_as: str | None = None,
        default_std: Number | None = None,
        description: str | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        self.set_description(description, )
        self._flags = _flags.Flags.from_kwargs(**kwargs, )
        self._default_std = _resolve_default_std(default_std, self._flags, )
        self._context = source.context or {}
        std_name_format = std_name_format or self._DEFAULT_STD_NAME_FORMAT
        std_description_format = std_description_format or self._DEFAULT_STD_DESCRIPTION_FORMAT
        #
        self.quantities = tuple(source.quantities)
        self.dynamic_equations = tuple(source.dynamic_equations)
        self.steady_equations = tuple(source.steady_equations)
        #
        # Create std_ parameters for unanticipated and measurement shocks
        if self._flags.is_stochastic:
            unanticipated_shocks = _quantities.generate_quantities_of_kind(self.quantities, _quantities.QuantityKind.UNANTICIPATED_SHOCK, )
            measurement_shocks = _quantities.generate_quantities_of_kind(self.quantities, _quantities.QuantityKind.MEASUREMENT_SHOCK, )
            #
            self.quantities += tuple(_generate_stds(
                unanticipated_shocks,
                _quantities.QuantityKind.UNANTICIPATED_STD,
                std_name_format,
                std_description_format,
                entry=len(self.quantities),
            ))
            self.quantities += tuple(_generate_stds(
                measurement_shocks,
                _quantities.QuantityKind.MEASUREMENT_STD,
                std_name_format,
                std_description_format,
                entry=len(self.quantities),
            ))
        #
        # Look up undeclared names; autodeclare these names or throw an error
        undeclared_names = _collect_undeclared_names(
            self.quantities,
            self.dynamic_equations + self.steady_equations,
        )
        if undeclared_names and autodeclare_as is None:
            raise _wrongdoings.IrisPieCritical([
                "These names are used in equations but not declared:",
                *undeclared_names,
                ])
        self.quantities += tuple(_generate_quantities_for_undeclared_names(
            undeclared_names, autodeclare_as, entry=len(self.quantities),
        ))
        #
        # Verify uniqueness of all quantity names
        _quantities.check_unique_names(self.quantities)
        #
        # Number of transition equations must equal number of transition variables
        _check_numbers_of_variables_equations(
            self.quantities,
            self.dynamic_equations,
            _quantities.QuantityKind.TRANSITION_VARIABLE,
            _equations.EquationKind.TRANSITION_EQUATION,
        )
        #
        # Number of measurement equations must equal number of measurement variables
        _check_numbers_of_variables_equations(
            self.quantities,
            self.dynamic_equations,
            _quantities.QuantityKind.MEASUREMENT_VARIABLE,
            _equations.EquationKind.MEASUREMENT_EQUATION,
        )
        #
        self.quantities = _quantities.reorder_by_kind(self.quantities, )
        self.dynamic_equations = _equations.reorder_by_kind(self.dynamic_equations, )
        self.steady_equations = _equations.reorder_by_kind(self.steady_equations, )
        _quantities.stamp_id(self.quantities, )
        _equations.stamp_id(self.dynamic_equations, )
        _equations.stamp_id(self.steady_equations, )
        #
        self.shock_qid_to_std_qid = (
            _create_shock_qid_to_std_qid(self.quantities, std_name_format, )
            if self._flags.is_stochastic
            else {}
        )
        #
        name_to_qid = _quantities.create_name_to_qid(self.quantities, )
        _equations.finalize_dynamic_equations(self.dynamic_equations, name_to_qid, )
        _equations.finalize_steady_equations(self.steady_equations, name_to_qid, )
        #
        if check_syntax:
            _check_syntax(self.dynamic_equations, self._context, )
            _check_syntax(self.steady_equations, self._context, )
        #
        self._min_shift, self._max_shift = None, None
        self._populate_min_max_shifts()
        #
        self.dynamic_descriptor = _descriptors.Descriptor(self.dynamic_equations, self.quantities, self._context, )
        self.steady_descriptor = _descriptors.Descriptor(self.steady_equations, self.quantities, self._context, )
        #
        # Create a function to evaluate LHS–RHS in dynamic equations
        self._plain_dynamic_equator = _equators.PlainEquator(
            _equations.generate_equations_of_kind(self.dynamic_equations, _PLAIN_EQUATOR_EQUATION, ),
            context=self._context,
        )
        #
        # Create a function to evaluate LHS–RHS in steady equations
        self._plain_steady_equator = _equators.PlainEquator(
            _equations.generate_equations_of_kind(self.steady_equations, _PLAIN_EQUATOR_EQUATION, ),
            context=self._context,
        )

    @property
    def num_transition_equations(self, /, ) -> int:
        """
        """
        return sum(
            1 for e in self.dynamic_equations
            if e.kind is _equations.EquationKind.TRANSITION_EQUATION
        )

    @property
    def num_measurement_equations(self, /, ) -> int:
        """
        """
        return sum(
            1 for e in self.dynamic_equations
            if e.kind is _equations.EquationKind.MEASUREMENT_EQUATION
        )

    def _populate_min_max_shifts(self, /, ) -> None:
        """
        """
        self._min_shift = _equations.get_min_shift_from_equations(
            self.dynamic_equations + self.steady_equations,
        )
        self._max_shift = _equations.get_max_shift_from_equations(
            self.dynamic_equations + self.steady_equations,
        )

    #]


def _check_syntax(equations, function_context, /, ):
    """
    Try all equations at once; if this fails, do equation by equation to catch the troublemakers
    """
    try:
        eval(_equations.create_equator_func_string(equations, ), )
    except:
        _catch_troublemakers(equations, function_context, )
    #]


def _catch_troublemakers(equations, function_context, /, ):
    """
    Catch the troublemakers
    """
    #[
    fail = [
        eqn.human for eqn in equations
        if not _success_creating_lambda(eqn, function_context)
    ]
    if fail:
        message = ["Syntax error in these equations"] + fail
        raise _wrongdoings.IrisPieCritical(message, )
    #]


def _success_creating_lambda(equation, function_context):
    """
    """
    #[
    try:
        eval(_equations.create_equator_func_string([equation.xtring]))
        return True
    except Exception as ex:
        return False
    #]


def _create_shock_qid_to_std_qid(
    quantities: Iterable[_quantities.Quantity],
    std_name_format: str,
    /,
) -> dict[int, int]:
    """
    """
    #[
    name_to_qid = _quantities.create_name_to_qid(quantities, )
    qid_to_name = _quantities.create_qid_to_name(quantities, )
    kind = _quantities.QuantityKind.STOCHASTIC_SHOCK
    all_shock_qids = tuple(_quantities.generate_qids_by_kind(quantities, kind))
    return {
        shock_qid: name_to_qid[std_name_format.format(qid_to_name[shock_qid], )]
        for shock_qid in all_shock_qids
    }
    #]


def _generate_stds(
    shocks: tuple[_quantities.Quantity, ...],
    std_kind: _quantities.QuantityKind,
    std_name_format: str,
    std_description_format: str,
    /,
    *,
    entry: int,
) -> tuple[_quantities.Quantity, ...]:
    """
    """
    #[
    return (
        _quantities.Quantity(
            id=None,
            human=std_name_format.format(shock.human, ),
            kind=std_kind,
            logly=None,
            description=std_description_format.format(shock.description or shock.human, ),
            entry=entry,
        )
        for shock in shocks
    )
    #]


def _collect_undeclared_names(
    quantities: tuple[_quantities.Quantity, ...],
    equations: tuple[_equations.Equation, ...],
    /,
) -> tuple[str, ...]:
    """
    """
    #[
    all_names = set(_quantities.generate_all_quantity_names(quantities, ))
    all_names_in_equations = set(_equations.generate_all_names_from_equations(equations, ))
    return tuple(all_names_in_equations - all_names)
    #]


def _generate_quantities_for_undeclared_names(
    undeclared_names: set[str],
    autodeclare_as: str | None,
    /,
    *,
    entry: int,
) -> tuple[_quantities.Quantity, ...] | NoReturn:
    """
    """
    #[
    if not undeclared_names:
        return tuple()
    autodeclare_kind = _quantities.QuantityKind.from_keyword(autodeclare_as, )
    return (
        _quantities.Quantity(
            id=None,
            human=name,
            kind=autodeclare_kind,
            logly=None,
            description=None,
            entry=entry,
        )
        for name in undeclared_names
    )
    #]


def _check_numbers_of_variables_equations(
    quantities: tuple[_quantities.Quantity, ...],
    equations: tuple[_equations.Equation, ...],
    quantity_kind: _quantities.QuantityKind,
    equation_kind: _equations.EquationKind,
    /,
) -> None:
    """
    """
    #[
    num_variables = _quantities.count_quantities_of_kind(quantities, quantity_kind, )
    num_equations = _equations.count_equations_of_kind(equations, equation_kind, )
    if num_variables != num_equations:
        raise _wrongdoings.IrisPieCritical([
            "Inconsistent numbers of variables and equations",
            f"Number of {quantity_kind.human.lower()}s: {num_variables}",
            f"Number of {equation_kind.human.lower()}s: {num_equations}",
        ])
    #]


def _resolve_default_std(
    custom_default_std: Number | None,
    flags: _flags.Flags,
) -> Number:
    """
    """
    #[
    if custom_default_std is not None:
        return custom_default_std
    elif flags.is_linear:
        return _DEFAULT_STD_LINEAR
    else:
        return _DEFAULT_STD_NONLINEAR
    #]




"""
Main time series class definition
"""


#[
from __future__ import annotations

from numbers import (Real, )
from collections.abc import (Iterable, Callable, )
from typing import (Self, Any, TypeAlias, NoReturn, )
from types import (EllipsisType, )
import numpy as _np
import scipy as _sp
import itertools as _it
import functools as _ft
import operator as _op
import copy as _cp

from ..conveniences import descriptions as _descriptions
from ..conveniences import copies as _copies
from ..conveniences import iterators as _iterators
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants
from .. import pages as _pages

from . import _temporal
from . import _filling
from . import _moving
from . import _conversions
from . import _hp
from . import _x13

from . import _indexing
from . import _plotly
from . import _views
from . import _functionalize

from ._temporal import __all__ as _diffs_cums__all__
from ._temporal import *

from ._filling import __all__ as _fillings__all__
from ._filling import *

from ._hp import __all__ as _hp__all__
from ._hp import *

from ._moving import __all__ as _moving__all__
from ._moving import *

from ._conversions import __all__ as _conversions__all__
from ._conversions import *
#]


_ELEMENTWISE_FUNCTIONS = {
    "log": _np.log,
    "exp": _np.exp,
    "sqrt": _np.sqrt,
    "abs": _np.abs,
    "sign": _np.sign,
    "sin": _np.sin,
    "cos": _np.cos,
    "tan": _np.tan,
    "round": _np.round,
    "logistic": _sp.special.expit,
    "normal_cdf": _sp.stats.norm.cdf,
    "normal_pdf": _sp.stats.norm.pdf,
}


# FIXME
FUNCTION_ADAPTATIONS_ELEMENTWISE = tuple(_ELEMENTWISE_FUNCTIONS.keys())
FUNCTION_ADAPTATIONS_NUMPY_APPLY = () # ("maximum", "minimum", "mean", "median", "abs", )
FUNCTION_ADAPTATIONS_BUILTINS = ("max", "min", )
FUNCTION_ADAPTATIONS = tuple(set(
    FUNCTION_ADAPTATIONS_ELEMENTWISE
    # + FUNCTION_ADAPTATIONS_NUMPY
    # + FUNCTION_ADAPTATIONS_BUILTINS
))


__all__ = (
    ("Series", "shift", )
    + _conversions__all__
    + _diffs_cums__all__
    + _fillings__all__
    + _hp__all__
    + _moving__all__
    + FUNCTION_ADAPTATIONS
)

Dates: TypeAlias = _dates.Dater | Iterable[_dates.Dater] | _dates.Ranger | EllipsisType | None
VariantsRequestType: TypeAlias = int | Iterable[int] | slice | None


def _get_date_positions(dates, base, num_periods, /, ):
    pos = tuple(_dates.date_index(dates, base))
    min_pos = _builtin_min((x for x in pos if x is not None), default=0)
    max_pos = _builtin_max((x for x in pos if x is not None), default=0)
    add_before = _builtin_max(-min_pos, 0)
    add_after = _builtin_max(max_pos - num_periods + 1, 0)
    pos_adjusted = [
        p + add_before if p is not None else None
        for p in pos
    ]
    return pos_adjusted, add_before, add_after


def _trim_decorate(func):
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        self.trim()
        return self
    return wrapper


@_pages.reference(
    path=("data_management", "time_series.md", ),
    categories={
        "constructor": "New time series",
        "conversion": "Time series frequency conversions",
        "filtering": "Univariate time series filters",
        "moving": "Moving window",
        "temporal_change": "Temporal change",
        "temporal_cumulation": "Temporal cumulation",
    },
)
class Series(
    _indexing.Inlay,
    _conversions.Inlay,
    _temporal.Inlay,
    _filling.Inlay,
    _hp.Inlay,
    _moving.Inlay,
    _x13.Inlay,
    _plotly.Inlay,

    _descriptions.DescriptionMixin,
    _views.ViewMixin,
    _copies.CopyMixin,
):
    r"""
················································································

Time series
============

The `Series` objects represent numerical time series, organized as rows of
observations stored in NumPy arrays and [date](dates.md)
stamped. A `Series` object can hold multiple
variants of the data, stored as mutliple columns.

················································································
    """
    #[

    __slots__ = (
        "start_date",
        "data",
        "data_type",
        "metadata",
        "_description", 
    )

    _numeric_format: str = "15g"
    _short_str_format: str = ">15"
    _date_str_format: str = ">12"
    _missing = _np.nan
    _missing_str: str = "⋅"
    _test_missing_period = staticmethod(lambda x: _np.all(_np.isnan(x)))

    @_pages.reference(
        category="constructor",
        call_name="Series",
    )
    def __init__(
        self,
        /,
        *,
        num_variants: int = 1,
        data_type: type = _np.float64,
        description: str = "",
        start_date: _dates.Dater | None = None,
        dates: Iterable[_dates.Dater] | None = None,
        frequency: _dates.Frequency | None = None,
        values: Any | None = None,
        func: Callable | None = None,
    ) -> None:
        """
················································································

==Create a new `Series` object==

```
self = Series(
    start_date=start_date,
    values=values,
)
```

```
self = Series(
    dates=dates,
    values=values,
)
```

### Input arguments ###


???+ input "start_date"

    The date of the first value in the `values`.

???+ input "values"

    Time series observations, supplied either as a tuple of values, or a
    NumPy array.

### Returns ###

???+ returns "self"

    The newly created `Series` object.

················································································
        """
        self.start_date = None
        self.data_type = data_type
        self.data = _np.full((0, num_variants), self._missing, dtype=self.data_type)
        self.metadata = {}
        self._description = description
        test = (x is not None for x in (start_date, dates, values, func))
        _SERIES_FACTORY.get(tuple(test), _invalid_constructor)(self, start_date=start_date, dates=dates, frequency=frequency, values=values, func=func, )

    def reset(self, /, ) -> None:
        self.__init__(
            num_variants=self.num_variants,
            data_type=self.data_type,
        )

    def _create_periods_of_missing_values(self, num_rows=0) -> _np.ndarray:
        """
        """
        return _np.full(
            (num_rows, self.shape[1], ),
            self._missing,
            dtype=self.data_type
        )

    @property
    @_pages.reference(category="property", )
    def shape(self, /, ) -> tuple[int, int]:
        """==Shape of time series data=="""
        return self.data.shape

    @property
    @_pages.reference(category="property", )
    def num_periods(self, /, ) -> int:
        """==Number of periods from the first to the last observation=="""
        return self.data.shape[0]

    @property
    @_pages.reference(category="property", )
    def num_variants(self, /, ) -> int:
        """==Number of variants (columns) within the `Series` object=="""
        return self.data.shape[1]

    @property
    def is_singleton(self, /, ) -> bool:
        """
        True for time series with only one variant
        """
        return _has_variants.is_singleton(self.num_variants, )

    @property
    @_pages.reference(category="property", )
    def span(self, ):
        """==Time span of the time series=="""
        return _dates.Ranger(self.start_date, self.end_date, ) if self.start_date else ()

    range = span

    @property
    @_pages.reference(category="property", )
    def from_to(self, ):
        """==Two-tuple with the start date and end date of the time series=="""
        return self.start_date, self.end_date

    @property
    @_pages.reference(category="property", )
    def dates(self, /, ) -> tuple[_dates.Dater, ...]:
        """==N-tuple with the dates from the start date to the end date of the time series=="""
        return tuple(self.range, )

    @property
    @_pages.reference(category="property", call_name="start_date", )
    def _start_date(self):
        """==Start date of the time series=="""
        raise NotImplementedError

    @property
    @_pages.reference(category="property", )
    def end_date(self):
        """==End date of the time series=="""
        return self.start_date + self.data.shape[0] - 1 if self.start_date else None

    @property
    @_pages.reference(category="property", )
    def frequency(self):
        """==Date frequency of the time series=="""
        return (
            self.start_date.frequency
            if self.start_date is not None
            else _dates.Frequency.UNKNOWN
        )

    @classmethod
    def from_dates_and_values(
        klass,
        dates: Dates,
        values: Any,
        **kwargs,
    ) -> Self:
        """
        """
        return klass(
            dates=dates,
            values=values,
            **kwargs,
        )

    @classmethod
    def from_start_date_and_values(
        klass,
        start_date: _dates.Dater,
        values: _np.ndarray | Iterable,
        **kwargs,
    ) -> Self:
        return klass(
            start_date=start_date,
            values=values,
            **kwargs,
        )

    @classmethod
    def from_dates_and_func(
        klass,
        dates: Iterable[_dates.Dater],
        func: Callable,
        **kwargs,
    ) -> Self:
        return klass(
            dates=dates,
            func=func,
            **kwargs,
        )

    @_trim_decorate
    def set_data(
        self,
        dates: Dates,
        data: Any | Series,
        variants: VariantsRequestType = None,
        /,
    ) -> None:
        dates = self._resolve_dates(dates, )
        is_empty_data = (
            data is None
            or (isinstance(data, Series) and data.size == 0)
        )
        if not dates and is_empty_data:
            return
        data = (
            _reshape_numpy_array(data, )
            if isinstance(data, _np.ndarray) else data
        )
        vids = self._resolve_variants(variants, )
        if not self.start_date:
            self.start_date = next(iter(dates), None, )
            self.data = self._create_periods_of_missing_values(num_rows=1, )
        pos, add_before, add_after = _get_date_positions(dates, self.start_date, self.shape[0], )
        self.data = self._create_expanded_data(add_before, add_after)
        if add_before:
            self.start_date -= add_before
        if hasattr(data, "get_data"):
            data = data.get_data(dates)
        #
        data_variants = _has_variants.iter_variants(data, )
        for c, d in zip(vids, data_variants, ):
            self.data[pos, c] = d

    def _get_data_and_recreate(
        self,
        *args,
    ) -> _np.ndarray:
        """
        """
        dates, pos, variants, expanded_data = self._resolve_dates_and_positions(*args, )
        data = expanded_data[_np.ix_(pos, variants)]
        num_variants = data.shape[1]
        new = Series(num_variants=num_variants, data_type=self.data_type)
        new.set_data(dates, data)
        return new

    def _resolve_dates_and_positions(
        self,
        dates: Dates,
        variants: VariantsRequestType = None,
        /,
    ) -> tuple[Iterable[_dates.Dater], Iterable[int], Iterable[int], _np.ndarray]:
        """
        """
        dates = [ t for t in self._resolve_dates(dates, ) ]
        variants = self._resolve_variants(variants)
        if not dates:
            dates = ()
            pos = ()
            data = self._create_periods_of_missing_values(num_rows=0, )[:, variants]
            return dates, pos, variants, data
        #
        base_date = self.start_date or _builtin_min(dates, )
        pos, add_before, add_after = _get_date_positions(dates, base_date, self.shape[0], )
        data = self._create_expanded_data(add_before, add_after, )
        if not isinstance(pos, Iterable):
            pos = (pos, )
        return dates, pos, variants, data

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> Self:
        """
        Alter (expand, shrink) the number of variants in this time series object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )

    def expand_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        """
        add = new_num - self.num_variants
        if add < 0:
            raise ValueError("Use shrink_num_variants to shrink the number of variants")
        data = _np.tile(self.data[:, (-1, )], (1, add, ))
        self.data = _np.hstack((self.data, data, ), )

    def shrink_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        """
        remove = self.num_variants - new_num
        if remove < 0:
            raise ValueError("Use expand_num_variants to expand the number of variants")
        self.data = self.data[:, :-remove]

    def get_data(
        self,
        dates = ...,
        *args,
    ) -> _np.ndarray:
        dates, pos, variants, expanded_data = self._resolve_dates_and_positions(dates, *args, )
        data = expanded_data[_np.ix_(pos, variants)]
        return data

    def get_data_variant(
        self,
        dates: Dates,
        variant: Real | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant<self.data.shape[1] else 0
        return self.get_data(dates, variant, )

    def get_data_from_to(
        self,
        from_to,
        *args,
    ) -> _np.ndarray:
        """
        """
        _, pos, variants, expanded_data \
            = self._resolve_dates_and_positions(from_to, *args, )
        from_pos, to_pos = pos[0], pos[-1]+1
        return expanded_data[from_pos:to_pos, variants]

    def get_data_variant_from_to(
        self,
        from_to: Iterable[_dates.Dater],
        variant: int | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant < self.data.shape[1] else 0
        return self.get_data_from_to(from_to, variant, )

    def extract_variants(
        self,
        variants,
        /,
    ) -> None:
        if not isinstance(variants, Iterable):
            variants = (variants, )
        else:
            variants = tuple(c for c in variants)
        self.data = self.data[:, variants]

    def set_start_date(
        self,
        new_start_date: _dates.Dater,
        /,
    ) -> Self:
        self.start_date = new_start_date
        return self

    def _resolve_dates(self, dates, ):
        """
        """
        if dates is None:
            # dates = []
            dates = ...
        if isinstance(dates, slice) and dates == slice(None, ):
            dates = ...
        if dates is ... and self.start_date is not None:
            dates = _dates.Ranger(None, None, )
        if dates is ... and self.start_date is None:
            dates = _dates.EmptyRanger()
        if hasattr(dates, "needs_resolve") and dates.needs_resolve:
            dates = dates.resolve(self, )
        return tuple(
            d.resolve(self) if hasattr(d, "needs_resolve") and d.needs_resolve else d
            for d in dates
        )

    def _resolve_variants(
        self,
        variants: VariantsRequestType,
        /,
    ) -> Iterable[int]:
        """
        Resolve variant request to an iterable of integers
        """
        if variants is None or variants is Ellipsis:
            variants = slice(None, )
        if isinstance(variants, slice):
            variants = range(*variants.indices(self.num_variants))
        if not isinstance(variants, Iterable):
            variants = (variants, )
        return variants

    def shift(
        self,
        by: int | str = -1,
        /,
    ) -> None:
        match by:
            case "yoy":
                self._shift_by_number(-self.frequency.value, )
            case "boy" | "soy":
                self._shift_to_soy()
            case "eopy":
                self._shift_to_eopy()
            case "tty":
                self._shift_to_tty()
            case _:
                self._shift_by_number(by, )

    def _shift_by_number(self, by: int, ) -> None:
        """
        Shift (lag, lead) start date by a number of periods
        """
        self.start_date = (
            self.start_date - by
            if self.start_date else self.start_date
        )

    def _shift_to_soy(self, ) -> None:
        new_data = self.get_data(
            t.create_soy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_to_tty(self, ) -> None:
        new_data = self.get_data(
            t.create_tty()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_to_eopy(self, ) -> Self:
        new_data = self.get_data(
            t.create_eopy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def hstack(self, *args):
        if not args:
            return self.copy()
        self_args = (self, *args, )
        if all(i.is_empty() for i in self_args):
            num_variants = sum(i.num_variants for i in self_args)
            return Series(num_variants=num_variants, )
        encompassing_span, *from_to = _dates.get_encompassing_span(self, *args, )
        new_data = self.get_data_from_to(from_to, )
        add_data = (
            x.get_data_from_to(from_to, )
            if hasattr(x, "get_data")
            else _create_data_variant_from_number(x, encompassing_span, self.data_type)
            for x in args
        )
        new_data = _np.hstack((new_data, *add_data))
        new = Series(num_variants=new_data.shape[1], )
        new.set_data(encompassing_span, new_data, )
        new.trim()
        return new

    def clip(
        self,
        /,
        new_start_date: _dates.Dater | None,
        new_end_date: _dates.Dater | None,
    ) -> None:
        if new_start_date is None or new_start_date < self.start_date:
            new_start_date = self.start_date
        if new_end_date is None or new_end_date > self.end_date:
            new_end_date = self.end_date
        if new_start_date == self.start_date and new_end_date == self.end_date:
            return
        self.data = self.get_data_from_to((new_start_date, new_end_date, ), )
        self.start_date = new_start_date

    def is_empty(self, ) -> bool:
        return not self.data.size

    def empty(
        self,
        *,
        num_variants: int | None = None
    ) -> None:
        num_variants = num_variants if num_variants is not None else self.num_variants
        self.data = _np.empty((0, num_variants), dtype=self.data_type)

    def overlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        self.set_data(other.range, other.data, )
        self.trim()

    def overlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        _broadcast_variants_if_needed(self, other, )
        self._LAY_METHOD_RESOLUTION[method]["overlay"](self, other, )

    def underlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        """
        """
        new_self = other.copy()
        new_self.overlay(self, )
        self._shallow_copy_data(new_self, )

    def underlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        _broadcast_variants_if_needed(self, other, )
        self._LAY_METHOD_RESOLUTION[method]["underlay"](self, other, )

    _LAY_METHOD_RESOLUTION = {
        "by_range": {"overlay": overlay_by_range, "underlay": underlay_by_range},
    }

    def redate(
        self,
        new_date: _dates.Dater,
        old_date: _dates.Dater | None = None,
        /,
    ) -> None:
        """
        """
        self.start_date = new_date \
            if old_date is None \
            else new_date - (old_date - self.start_date)

    def _shallow_copy_data(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        for n in ("start_date", "data", "data_type", ):
            setattr(self, n, getattr(other, n, ))

    def __or__(self, other):
        """
        Implement the | operator
        """
        return self.hstack(other)

    def trim(self):
        if self.data.size == 0:
            self.reset()
            return self
        num_leading = _get_num_leading_missing_rows(self.data, self._test_missing_period)
        if num_leading == self.data.shape[0]:
            self.reset()
            return self
        num_trailing = _get_num_leading_missing_rows(self.data[::-1], self._test_missing_period)
        if not num_leading and not num_trailing:
            return self
        slice_from = num_leading or None
        slice_to = -num_trailing if num_trailing else None
        self.data = self.data[slice(slice_from, slice_to), ...]
        if slice_from:
            self.start_date += int(slice_from)
        return self

    def _create_expanded_data(self, add_before, add_after):
        return _np.pad(
            self.data, ((add_before, add_after), (0, 0)),
            mode="constant", constant_values=self._missing
        )

    def _check_data_shape(self, data, /, ):
        if data.shape[1] != self.data.shape[1]:
            raise Exception("Time series data being assigned must preserve the number of variants")

    def __bool__(self, /, ):
        """
        """
        return self.data.size > 0

    def __neg__(self):
        """
        -self
        """
        output = self.copy()
        output.data = -output.data
        return output

    def __pos__(self):
        """
        +self
        """
        return self.copy()

    def __add__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.add, )

    def __radd__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__radd__(other))

    def __mul__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.mul, )

    def __rmul__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__rmul__(other))

    def __sub__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.sub, )

    def __rsub__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rsub__(other))

    def __pow__(self, other):
        """
        self ** other
        """
        # FIXME: 1 ** numpy.nan -> 1 !!!!!
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.pow, )

    def __rpow__(self, other):
        """
        other ** self
        """
        return self.apply(lambda data: data.__rpow__(other))

    def __truediv__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.truediv, )

    def __rtruediv__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rtruediv__(other))

    def __floordiv__(self, other):
        """
        self // other
        """
        return self._binop(other, _op.floordiv, )

    def __rfloordiv__(self, other):
        """
        other // self
        """
        return self.apply(lambda data: data.__rfloordiv__(other))

    def __mod__(self, other):
        """
        self % other
        """
        return self._binop(other, _op.mod, )

    def __rmod__(self, other, /, ):
        """
        other % self
        """
        return self.apply(lambda data: data.__rmod__(other))

    for n in ["gt", "lt", "ge", "le", "eq", "ne", ]:
        exec(f"def __{n}__(self, other): return self._binop(other, _op.{n}, )", )

    def apply(self, func, /, *args, **kwargs, ):
        new_data = func(self.data, *args, **kwargs, )
        axis = kwargs.get("axis", None, )
        if new_data.shape == self.data.shape:
            new = self.copy()
            new._replace_data(new_data, )
            return new
        elif (axis is None or axis == 1) and new_data.shape == (self.data.shape[0], ):
            new = self.copy()
            new.data = new_data.reshape(self.data.shape[0], 1, )
            new.trim()
            return new
        elif axis is None and new_data.shape == ():
            return new_data
        elif axis == 0 and new_data.shape == (self.data.shape[1], ):
            return new_data
        else:
            raise _wrongdoings.IrisPieError(
                "Function applied on a time series resulted"
                " in a data array with an unexpected shape"
            )

    def _binop(self, other, func, /, new=None, ):
        if not isinstance(other, type(self)):
            return self.apply(lambda data: func(data, other))
        # FIXME: empty encompassing range
        _, *from_to = _dates.get_encompassing_span(self, other)
        self_data = self.get_data_from_to(from_to, )
        other_data = other.get_data_from_to(from_to, )
        new_data = func(self_data, other_data)
        new = Series(num_variants=new_data.shape[1], ) if new is None else new
        new._replace_start_date_and_values(from_to[0], new_data, )
        return new

    def _broadcast_variants(self, num_variants, /, ) -> None:
        """
        """
        if self.data.shape[1] == num_variants:
            return
        if self.data.shape[1] == 1:
            self.data = _np.repeat(self.data, num_variants, axis=1, )
            return
        raise _wrongdoings.IrisPieError("Cannot broadcast variants")

    def _replace_data(
        self,
        new_values,
        /,
    ) -> None:
        """
        """
        self.data = new_values
        self.trim()

    def _replace_start_date_and_values(
        self,
        new_start_date,
        new_values,
        /,
    ) -> None:
        """
        """
        self.start_date = new_start_date
        self._replace_data(new_values, )

    def iter_dates_values(self, /, unpack_singleton=True, ):
        """
        Default iterator is line by line, yielding a tuple of (date, values)
        """
        def _unpack_singleton_data_row(data_list: list[Real], /, ):
            return _has_variants.unpack_singleton(data_list, True, )
        def _keep_data_row(data_list: list[Real], /, ):
            return data_list
        data_row_func = (
            _unpack_singleton_data_row
            if self.is_singleton and unpack_singleton
            else _keep_data_row
        )
        for date, data_row in zip(self.range, self.data, ):
            yield date, data_row_func(data_row.tolist(), )

    def iter_variants(self, /, ) -> Iterator[Self]:
        """
        """
        for data in self.iter_data_variants_from_to(..., ):
            new = self.copy()
            new._replace_data(data, )
            yield new

    def iter_own_data_variants_from_to(self, from_to, /, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        data_from_to = self.data if from_to == ... else self.get_data_from_to(from_to, )
        return iter(data_from_to.T, )

    def iter_data_variants_from_to(self, from_to, /, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        return _iterators.exhaust_then_last(self.iter_own_data_variants_from_to(from_to, ), )

    def logistic(self, /, ) -> Self:
        """
        """
        self.data = 1 / (1 + _np.exp(-self.data))

    for n in FUNCTION_ADAPTATIONS_ELEMENTWISE:
        exec(f"def {n}(self, *args, **kwargs, ): self.data = _ELEMENTWISE_FUNCTIONS['{n}'](self.data, *args, **kwargs, )")

    #]


Series.from_dates_and_data = _wrongdoings.obsolete(Series.from_dates_and_values)
Series.from_start_date_and_data = _wrongdoings.obsolete(Series.from_start_date_and_values)


def _get_num_leading_missing_rows(data, test_missing_period, /, ):
    try:
        num = next(
            i for i, period_data in enumerate(data)
            if not test_missing_period(period_data)
        )
    except StopIteration:
        num = data.shape[0]
    return num


# def apply(x, func: Callable, /, *args, **kwargs) -> object:
    # if isinstance(x, Series):
        # return x.apply(func, *args, **kwargs)
    # else:
        # return func(x, *args, **kwargs)


def hstack(first, *args) -> Self:
    return first.hstack(*args)


def _create_data_variant_from_number(
    number: Real,
    range: _dates.Ranger,
    data_type: type,
    /,
) -> _np.ndarray:
    return _np.full((len(range), 1), number, dtype=data_type)


for n in FUNCTION_ADAPTATIONS_ELEMENTWISE:
    exec(
        f"@_ft.singledispatch\n"
        f"def {n}(x, *args, **kwargs):\n"
        f"    return _ELEMENTWISE_FUNCTIONS['{n}'](x, *args, **kwargs)\n"
    )
    exec(
        f"@{n}.register(Series, )\n"
        f"def _(x, *args, **kwargs):\n"
        f"    new = x.copy()\n"
        f"    new.{n}(*args, **kwargs)\n"
        f"    return new\n"
    )


for n in FUNCTION_ADAPTATIONS_BUILTINS:
    exec(
        f"_builtin_{n} = {n}"
    )
#     exec(
#         f"@_ft.singledispatch"
#         f"\n"
#         f"def {n}(x, *args, **kwargs): "
#         f"return _builtin_{n}(x, *args, **kwargs)"
#     )
#     exec(
#         f"@{n}.register(Series, )"
#         f"\n"
#         f"def {n}(x, *args, **kwargs): "
#         f"return x._{n}_(*args, **kwargs)"
#     )


def _from_dates_and_values(
    self,
    dates: Iterable[_dates.Dater] | str,
    values: _np.ndarray | Iterable,
    frequency: _dates.Frequency | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    # dates = _dates.ensure_date_tuple(dates, frequency=frequency, )
    self.set_data(dates, values, )
    #]


def _from_dates_and_func(
    self,
    dates: Iterable[_dates.Dater] | str,
    func: Callable,
    frequency: _dates.Frequency | None = None,
    **kwargs,
) -> Self:
    """
    Create a new time series from dates and a function
    """
    #[
    # dates = _dates.ensure_date_tuple(dates, frequency=frequency, )
    data = [
        [func() for j in range(self.num_variants)]
        for i in range(len(dates))
    ]
    data = _np.array(data, dtype=self.data_type)
    self.set_data(dates, data)
    #]


def _from_start_date_and_values(
    self,
    start_date: _dates.Dater,
    values: _np.ndarray | Iterable,
    frequency: _dates.Frequency | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    # start_date = _dates.ensure_date_tuple(start_date, frequency=frequency, )[0]
    self.start_date = start_date
    if isinstance(values, _np.ndarray):
        values = _reshape_numpy_array(values, )
    else:
        values = _has_variants.iter_variants(values, )
        values = _np.column_stack(tuple(
            v for v, _ in zip(values, range(self.num_variants), )
        ))
    values = values.astype(self.data_type, )
    self.data = values
    self.trim()
    #]


def _reshape_numpy_array(values: _np.ndarray, /, ) -> _np.ndarray:
    """
    """
    #[
    return (
        values.reshape(values.shape[0], -1)
        if values.ndim >= 2 else values.reshape(-1, 1)
    )
    #]


def _broadcast_variants_if_needed(
    self: Series,
    other: Series,
    /,
) -> tuple[Series, Series]:
    """
    """
    #[
    if self.num_variants == other.num_variants:
        return self, other
    #
    if self.num_variants == 1:
        return self._broadcast_variants(other.num_variants, ), other
    #
    if other.num_variants == 1:
        return self, other._broadcast_variants(self.num_variants, )
    #
    raise _wrongdoings.IrisPieError("Cannot broadcast time series variants")
    #]


def _invalid_constructor(
    self,
    **kwargs,
) -> NoReturn:
    raise _wrongdoings.IrisPieError("Invalid Series object constructor")


_SERIES_FACTORY = {
    (False, False, False, False): lambda self, **kwargs: None,
    (True, False, True, False): _from_start_date_and_values,
    (False, True, True, False): _from_dates_and_values,
    (False, True, False, True): _from_dates_and_func,
}


def shift(
    self,
    by: int | str = -1,
) -> _series.Series:
    """
    """
    #[
    new = self.copy()
    new.shift(by)
    return new
    #]


for n in ("underlay", "overlay", "redate", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )




"""
"""


#[
from __future__ import annotations

from ..conveniences import views as _views
#]


class ViewMixin(_views.ViewMixin, ):
    """
    """
    #[

    def _get_first_line_view(self, /, ):
        """
        """
        shape = self.data.shape
        range_str = f"{self.start_date}»{self.end_date}" if self.start_date is not None else "None"
        return f"Series {range_str} {shape[0]}×{shape[1]}"

    def _get_content_view_(self, /, ):
        """
        """
        return tuple(
            _get_series_row_str_(date, data_row, self._date_str_format, self._numeric_format, self._missing_str) 
            for row, (date, data_row) in enumerate(zip(self.range, self.data))
        )

    def _get_short_row_(self):
        """
        Create string representing rows skipped in the short view of times series
        """
        shape = self.data.shape
        date_str = ("{:"+self._date_str_format+"}").format(_views._VERTICAL_ELLIPSIS)
        value_format = "{:"+self._short_str_format+"}"
        data_str = "".join([value_format.format(_views._VERTICAL_ELLIPSIS)] * shape[1])
        return date_str + data_str

    #]


def _get_series_row_str_(date, data, date_str_format, numeric_format, missing_str: str):
    """
    String representing one row of time series data including the date
    """
    #[
    date_str = ("{:"+date_str_format+"}").format(date)
    value_format = "{:"+numeric_format+"}"
    data_str = "".join([value_format.format(v) for v in data])
    data_str = data_str.replace("nan", "{:>3}".format(missing_str))
    return date_str + data_str
    #]




"""
"""


#[
from __future__ import annotations

from typing import Any
from numbers import Real
import numpy as _np
import itertools as _it
import wlogging as _wl

from .. import pages as _pages
from .. import wrongdoings as _wrongdoings
from ..databoxes import main as _databoxes
from ..plans import main as _plans
from ..explanatories import main as _explanatories
from ..dataslates import main as _dataslates
#]


_LOGGER_NAME = "Sequential.simulate"
_dataslate_constructor = _dataslates.Dataslate.from_databox_for_slatable


class Inlay:
    """
    """
    #[

    @_pages.reference(category="simulation", )
    def simulate(
        self,
        input_db: _databoxes.Databox,
        span: Iterable[Dater],
        /,
        plan: _plans.Plan | None = None,
        prepend_input: bool = True,
        target_databox: _databoxes.Databox | None = None,
        when_nonfinite: Literal["error", "warning", "silent", ] = "warning",
        execution_order: Literal["dates_equations", "equations_dates", ] = "dates_equations",
        num_variants: int | None = None,
        remove_initial: bool = True,
        remove_terminal: bool = True,
        shocks_from_databox: bool = True,
        logging_level: int = _wl.INFO,
    ) -> tuple[_databoxes.Databox, dict[str, Any]]:
        """
......................................................................

==Simulate sequential model==

```
output_db, info = self.simulate(
    input_db, span,
    /,
    prepend_input=True,
    plan=None,
    target_databox=None,
    when_nonfinite="warning",
    num_variants=None,
    remove_initial=True,
    remove_terminal=True,
)
```

Simulate a `Sequential` model, `self`, on a time `span`, period by period,
equation by equation. The `simulate` function does not reorder the
equations; if needed, this must be done by running `reorder` before
simulating the model.


### Input arguments ###


???+ input "self"
    `Sequential` model that will be simulated.

???+ input "input_db"
    Input databox (a `Databox` object) with all the necessary initial
    conditions (initial lags) for the LHS variables, and all the values for
    the RHS variables needed to simulate `self` on the time `span`.

???+ input "plan"
    `PlanSimulate` object with a simulation plan, i.e. information about
    which LHS variables to exogenize at which dates. If `plan=None`, no
    simulation plan is imposed on the simulation.


### Returns ###

???+ returns "output_db"

    Output databox with the simulated time series for the LHS variables.

???+ returns "info"

    Information about the simulation; `info` is a dict with the following
    items.

......................................................................
        """
        logger = _wl.get_colored_logger(_LOGGER_NAME, level=logging_level, )
        num_variants = self.num_variants if num_variants is None else num_variants
        base_dates = tuple(span, )
        #
        work_db = input_db.shallow()
        if not shocks_from_databox:
            # REFACTOR: Create `quantities` in Sequential.invariant and
            # create a `get_names` method
            shock_names = self.residual_names
            work_db.remove(shock_names, strict_names=False, )
        #
        extra_databox_names = None
        if plan is not None:
            plan.check_consistency(self, base_dates, )
            extra_databox_names = plan.get_databox_names()
        #
        dataslate = _dataslate_constructor(
            self, work_db, base_dates,
            num_variants=self.num_variants,
            extra_databox_names=extra_databox_names,
        )
        #
        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            dataslate.iter_variants(),
        )
        #
        #=======================================================================
        # Main loop over variants
        for vid, model_v, dataslate_v in zipped:
            model_v._simulate_v(
                dataslate_v,
                logger,
                plan=plan,
                when_nonfinite=when_nonfinite,
                execution_order=execution_order,
            )
        #=======================================================================
        #
        # Remove initial and terminal condition data (all lags and leads
        # before and after the simulation span)
        if remove_terminal:
            dataslate.remove_terminal()
        if remove_initial:
            dataslate.remove_initial()
        #
        # Convert all variants of the dataslate to a databox
        output_db = dataslate.to_databox()
        if prepend_input:
            output_db.prepend(input_db, base_dates[0]-1, )
        #
        # Add to custom databox
        if target_databox is not None:
            output_db = target_databox | output_db
        #
        info = {}
        #
        return output_db, info

    def _simulate_v(
        self,
        ds: _dataslates.Dataslate,
        logger: _wl.Logger,
        /,
        *,
        plan,
        when_nonfinite,
        execution_order,
    ) -> None:
        """
        """
        when_nonfinite_stream = \
            _wrongdoings.STREAM_FACTORY[when_nonfinite] \
            ("These simulated data point(s) are non-finite values:", )
        #
        name_to_row = ds.create_name_to_row()
        base_columns = ds.base_columns
        first_base_column = base_columns[0] if base_columns else None
        working_data = ds.get_data_variant(0, )
        columns_dates = ( (c, ds.dates[c]) for c in base_columns )
        execution_iterator_creator = _CREATE_EXECUTION_ITERATOR[execution_order]
        columns_dates_equations = execution_iterator_creator(columns_dates, self.iter_equations(), )
        for (column, date), equation in columns_dates_equations:
            transform = _get_transform(plan, equation, date)
            is_exogenized, implied_value = _is_exogenized(
                equation.lhs_name,
                transform,
                working_data,
                column,
                name_to_row,
            )
            #
            simulation_function = equation.simulate if not is_exogenized else equation.exogenize
            info = simulation_function(working_data, column, implied_value, )
            logger.debug(
                f"{simulation_function.__name__}"
                f" {equation.lhs_name}[{date}]={info['lhs_value']}"
                f" (equation.residual_name[{date}]={info['residual_value']})"
            )
            #
            _catch_nonfinite(
                when_nonfinite_stream,
                info["is_finite"],
                info["simulated_name"],
                date,
            )
        when_nonfinite_stream._raise()

    #]


def _get_transform(
    plan: _plans.Plan | None,
    equation: _explanatories.Explanatory,
    date: _dates.Dater,
) -> _plans.Transform | None:
    """
    """
    return (
        plan.get_exogenized_point(equation.lhs_name, date, )
        if plan and not equation.is_identity
        else None
    )


def _is_exogenized(
    lhs_name: str,
    transform: _plans.Transform | None,
    data: _np.ndarray,
    data_column: int,
    name_to_row: dict[str, int],
) -> tuple[bool, Real | None]:
    """
    """
    #[
    if transform is None:
        return False, None
    #
    lhs_name_row = name_to_row[lhs_name]
    values_before = data[lhs_name_row, :data_column]
    values_after_inclusive = data[lhs_name_row, data_column:]
    #
    transform_name = transform.resolve_databox_name(lhs_name, )
    transform_row = name_to_row.get(transform_name, None, )
    transform_values_after = data[transform_row, data_column:] if transform_row is not None else None
    implied_value = transform.eval_exogenized(transform_values_after, values_before, values_after_inclusive, )
    if transform.when_data and _np.isnan(implied_value):
        return False, None
    return True, implied_value
    #]


def _catch_nonfinite(
    stream: _wrongdoings.Stream,
    is_finite: bool | _np.ndarray,
    simulated_name: str,
    date: Dater,
) -> None:
    """
    """
    #[
    if is_finite.all():
        return
    message = f"{simulated_name}[{date}]"
    stream.add(message, )
    #]


def _iter_dates_equations(columns_dates, equations, ) -> Iterator[tuple[int, _dates.Dater], _explanatories.Explanatory]:
    return _it.product(columns_dates, equations, )


def _iter_equations_dates(columns_dates, equations, ) -> Iterator[tuple[int, _dates.Dater], _explanatories.Explanatory]:
    return _swap_product(_it.product(equations, columns_dates, ), )


def _swap_product(iterator: Iterable[tuple[Any, Any]], ) -> Iterable[tuple[Any, Any]]:
    return ((b, a) for a, b in iterator)


_CREATE_EXECUTION_ITERATOR = {
    "dates_equations": _iter_dates_equations,
    "equations_dates": _iter_equations_dates,
}


